/* Copyright (C) 2017 Centro de Computacao Cientifica e Software Livre
 * Departamento de Informatica - Universidade Federal do Parana - C3SL/UFPR
 *
 * This file is part of simmc-agent
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301,
 * USA.
 */
#include <agent/get.h>
#include <string>

/*---------------------------------------------------------------------
 * Function: check_update
 * Summary: Get agent version from client and web service to check if
 * needed update agent.
 * Method: GET
 * Get agent agent version from webService.
 * Make download from agent
 */

int check_update() {
    Agent agent;
    Conf conf;
    Proxy proxy;
    Rotate rotate;
    Json::Value data;
    std::string version;

    /* Set agent prefix */
    agent.setPrefix(__FILE__, 3);

    /* Set config file */
    agent.setDefaultConf("/conf/datasid-conf.json");

    /* run common.cpp to make parse from config and proxy file */
    if (common(&data, &agent, &conf, &proxy)) {
        rotate_log_file(&agent, &data);
        std::fstream logFile(agent.getLogFile().c_str(), std::fstream::app);
        logFile << __DATE__ << " - Agent started." << std::endl;
        logFile << __DATE__ << " - Start of update process execution."
                    << std::endl;

        /* If proxy is setted */
        if (proxy.proxyFileIsSet()) {
            auto p = cpr::Get(cpr::Url{conf.getWebService() + "/api/v1/get"},
                  cpr::Proxies{{"http", proxy.getHost()}},
                  cpr::Parameters{{"agent", "version"}},
                  cpr::Header{{"content-type", "application/json"}});

            /* Check error with connection */
            if (p.status_code == 502) {
                agent.setMessage(__DATE__ +
                std::string(" - ERROR: The proxy server received an") +
                std::string(" invalid response from server."));

                logFile << agent.getMessage() << std::endl;
                data["error"]["proxy_error"] = agent.getMessage();
            }

            version = p.header["version"];

        } else {
            /* Contact webservice server to get client version. */
            auto p = cpr::Get(cpr::Url{conf.getWebService() + "/api/v1/get"},
                      cpr::Parameters{{"agent", "version"}},
                      cpr::Header{{"content-type", "application/json"}});

            /* Check error with connection */
            if (p.status_code == 502) {
                agent.setMessage(__DATE__ +
                std::string(" - ERROR: The proxy server received an") +
                std::string(" invalid response from server."));

                logFile << agent.getMessage()  << std::endl;
                data["error"]["proxy_error"] = agent.getMessage();
            }

            version = p.header["version"];
        }

        /* Case version are different, then make the download */
        if (compare_version(&data, &agent, version, &logFile)) {
            /* Case not download, return false */
            if (download(conf, &agent, &data, &logFile)) {
                agent.setMessage(__DATE__ +
                std::string(" - SUCCESS: Update process terminated ") +
                std::string("successfully."));

                logFile << agent.getMessage()  << std::endl;
                data["success"]["update_process"] = agent.getMessage();
            } else {
                agent.setMessage(__DATE__ +
                std::string(" - ERROR: Failed to update agent."));

                logFile << agent.getMessage() << std::endl;
                data["error"]["failed_update_process"] = agent.getMessage();
            }
        } else {
            agent.setMessage(__DATE__ +
            std::string(" - ERROR: Failed to update agent."));

            logFile << agent.getMessage() << std::endl;
            data["error"]["failed_update_process"] = agent.getMessage();
        }
        logFile.close();
    }

    #ifdef UPDATE_JSON
        Json::StyledWriter styledWriter;
        if (!data.isNull())
             std::cout << styledWriter.write(data) <<  std::endl;
        else
            std::cout << "Empty update object" << std::endl;
    #endif

    return(0);
}

/*---------------------------------------------------------------------
 * Function: compare_version
 * Summary: Compare client agent version with web service version.
 * If case are different and web service version not a empty string,
 * return true.
 */

bool compare_version(Json::Value* data, Agent *agent, std::string version,
                        std::fstream* logFile) {
    if (!strcmp(version.c_str(), "")) {
        agent->setMessage(__DATE__ +
        std::string(" - ERROR: Server returned an empty version string."));

        *logFile << agent->getMessage() << std::endl;
        (*data)["error"]["empty_string"] = agent->getMessage();

        return false;
    } else if (strcmp(version.c_str(), agent->getVersion().c_str())) {
        agent->setMessage(__DATE__ +
        std::string(" - WARNING: Client version out-of-date.") +
            std::string(" Update client."));

        *logFile << agent->getMessage() << std::endl;
        (*data)["data"]["client_version"] = agent->getMessage();

        return true;
    } else {
        agent->setMessage(__DATE__ +
        std::string(" - WARNING: Client is already updated."));

        *logFile << agent->getMessage() << std::endl;
        (*data)["data"]["client_agent"] = agent->getMessage();
        return false;
    }
}

/*---------------------------------------------------------------------
 * Function: copy_file_to_agent_home
 * Summary: Copy the agent zip file from tmp directory to opt directory.
 */

bool copy_zip_file_to_destiny_directory(Json::Value* data, std::fstream*
                logFile, Agent* agent, boost::filesystem::path source) {
    boost::filesystem::path destination = agent->getDirInstall();
    boost::filesystem::create_directory(destination);

    try {
        for (boost::filesystem::directory_iterator file(source);
        file != boost::filesystem::directory_iterator(); ++file) {
            boost::filesystem::path current(file->path());
            if (!boost::filesystem::is_directory(current)) {
                // Found file: Copy
                boost::filesystem::copy_file(current,
                    destination / current.filename());
            }
        }
    } catch (const boost::filesystem::filesystem_error& e) {
        std::string message = __DATE__ + std::string(" - ERROR: Failed to ")
            + std::string("copy the agent zipfile from tmp directory") +
            std::string(" to destiny directory.");

        *logFile << message << std::endl;
        (*data)["error"]["failed_to_copy_agent_zip"] = message;
        return false;
    }
    return true;
}

/*---------------------------------------------------------------------
 * Function: download
 * Summary: Make agent download.
 */

bool download(Conf conf, Agent* agent, Json::Value* data,
                std::fstream* logFile) {
    Curl curl;
    if (curl.downloadAgent(data, conf.getWebService(),
                            logFile, agent))
        return true;
    return false;
}
