/* Copyright (C) 2017 Centro de Computacao Cientifica e Software Livre
 * Departamento de Informatica - Universidade Federal do Parana - C3SL/UFPR
 *
 * This file is part of simmc-agent
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301,
 * USA.
 */
#include <agent/linux/common.h>
#include <string>

/*---------------------------------------------------------------------
 * Function: file_exist
 * Summary:  Check if a file exist.
 */

bool file_exist(const std::string& file) {
    std::ifstream myFile(file.c_str(), std::ifstream::in);
    if (myFile.is_open()) {
        myFile.close();
        return true;
    }
    myFile.close();
    return false;
}

/*---------------------------------------------------------------------
 * Function: create_directories
 * Summary:  Create "data" and "log" directories if needed.
 * c_str() - converte c++ string to c string.
 */

void create_directories(Conf* conf) {
    mkdir(conf->getLogDir().c_str(), 0700);
    mkdir(conf->getDataDir().c_str(), 0700);
}

/*---------------------------------------------------------------------
 * Function: check_version_file_exist
 * Summary:  Check if version file exist and set version variable from agent object.
 * Otherwise, setted 0.0.0.
 */

bool check_version_file_exist(Agent* agent) {
    std::string versionFile = agent->getConfDir() + "/version.json";
    std::ifstream myFile(versionFile.c_str(), std::ifstream::in);

    if (myFile.is_open()) {
        std::string version;
        myFile >> version;
        agent->setVersion(version);
        myFile.close();
        return true;
    }
    return false;
}

/*---------------------------------------------------------------------
 * Function: common
 * Summary:  Make parse from config and proxy file.
 */

bool common(Json::Value* data, Agent* agent, Conf* conf,
                    Proxy* proxy) {
    try {
        agent->prefixISet();
    } catch (const char* error) {
        agent->setMessage(__DATE__ + *error);
        (*data)["error"]["prefix_not_set"] =  agent->getMessage();
        return false;
    }

    /* Read config file and declare only valid variables */

    if (!file_exist(agent->getDefaultConf())) {
        agent->setMessage(__DATE__ +
            std::string(" - ERROR: Config file not found."));

        (*data)["error"]["config_file_not_found"] = agent->getMessage();
        return false;
    } else {
       if (!parse_config_file(agent, conf, data)) {
            return false;
        }
    }

    /* Read proxy file and declare only valid variables */

    if (!file_exist(conf->getProxyConf())) {
        agent->setMessage(__DATE__ +
            std::string(" - WARNING: Proxy file not found."));

        (*data)["warning"]["proxy_file_not_found"] =  agent->getMessage();
    } else {
        if (!parse_proxy_file(conf, proxy, data)) {
            return false;
        }
    }

    create_directories(conf);

    agent->setConfDir("/conf");
    agent->setLogFile(conf->getLogDir(), "/datasid-agent.log");

    if (!check_version_file_exist(agent))
        agent->setVersion("0.0.0");

    if (conf->getWebService().size() == 0) {
        agent->setMessage(__DATE__ +
            std::string(" - ERROR: Web service url not configured."));

        (*data)["error"]["web_service_url_not_found"] = agent->getMessage();
        return false;
    }

    return true;
}
