/* Copyright (C) 2004-2017 Centro de Computacao Cientifica e Software Livre
 * Departamento de Informatica - Universidade Federal do Parana - C3SL/UFPR
 *
 * This file is part of simmc-agent
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301,
 * USA.
 */
#include <agent/linux/curl.h>
#include <string>

/*---------------------------------------------------------------------
 * get Functions
 */

std::string Curl::getData() const {
    return this->curlBuffer;
}

/*---------------------------------------------------------------------
 * Method: downloadAgent
 * Summary: Contact webservice server to update the agent
 * Without proxy connection. Not download file from https protocol.
 * Method: get
 */

bool Curl::downloadAgent(Json::Value* data, std::string url,
                            std::fstream* logFile, Agent* agent) {
    Unzip unzip;
    curl = curl_easy_init();
    if (!curl) {
        agent->setMessage(__DATE__ + std::string(" - ERROR: Curl ") +
            std::string(" did not initialize!"));

        *logFile << agent->getMessage() << std::endl;
        (*data)["error"]["curl_did_not_initialize"] = agent->getMessage();

        return false;
    }

    boost::system::error_code ec;
    boost::filesystem::path test_root(boost::filesystem::unique_path(
        boost::filesystem::temp_directory_path(ec) / "Agent%%%%-%%%%-%%%%"));

    if (!boost::filesystem::create_directory(test_root, ec) || ec) {
        agent->setMessage(__DATE__ + std::string(" - ERROR: Failed to create ")
                + std::string("a temporary directory."));

        *logFile << agent->getMessage() << std::endl;
        (*data)["error"]["create_temporary_directory"] = agent->getMessage();

        return false;
    }

    /* Create the agent zip tmp file into test_root */
    agent->setZipName("agent.zip");
    boost::filesystem::path test_file(test_root / agent->getZipName());

    FILE* fp = fopen(test_file.c_str(), "wb");

    if (!fp) {
        agent->setMessage(__DATE__ + std::string(" - ERROR: Failed to ") +
                    std::string("create file on the disk."));

        *logFile << agent->getMessage() << std::endl;
        (*data)["error"]["failed_to_create_file_on_the_disk"] =
                        agent->getMessage();

        return false;
    }

    curl_easy_setopt(curl, CURLOPT_URL, url.c_str());
    curl_easy_setopt(curl, CURLOPT_WRITEDATA, fp);
    curl_easy_setopt(curl, CURLOPT_WRITEFUNCTION, &Curl::callBackDownload);
    curl_easy_setopt(curl, CURLOPT_FOLLOWLOCATION, 1);
    curl_easy_setopt(curl, CURLOPT_VERBOSE, 1L);
    curl_easy_setopt(curl, CURLOPT_FAILONERROR, true);
    response = curl_easy_perform(curl);
    curl_easy_getinfo(curl, CURLINFO_RESPONSE_CODE, &httpCode);
    curl_easy_cleanup(curl);

    if (httpCode == 302) {
        agent->setMessage(__DATE__ + std::string(" - ERROR: Protocol HTTPS ") +
                    std::string("not supported or disabled in libcurl."));

        *logFile << agent->getMessage() << std::endl;
        (*data)["error"]["protocol_https_not_supported"] = agent->getMessage();

        return false;
    }

    if (httpCode == 504) {
        agent->setMessage(__DATE__ +
        std::string(" - ERROR: Time out."));

        *logFile << agent->getMessage() << std::endl;
        (*data)["error"]["time_out"] = agent->getMessage();

        return false;
    }

    fclose(fp);
    agent->setDirInstall("/opt/agentc3sl");
    bool err = copy_zip_file_to_destiny_directory(data, logFile, agent,
                                                    test_root);

    // remove agent tmp directory
    boost::uintmax_t res = boost::filesystem::remove_all(test_root);
    if (res < 2) {
        agent->setMessage(__DATE__ + std::string(" - WARNING: Failed ") +
                    std::string("to remove the agent tmp directory."));

        *logFile << agent->getMessage() << std::endl;
        (*data)["warning"]["failed_to_remove_agent_tmp_drectory"] =
                            agent->getMessage();
    }

    // case error to move agent zip file from tmp to agent home
    if (!err)
        return false;

    unzip.setZipFileName((agent->getDirInstall() + std::string("/")
                                    + agent->getZipName()).c_str());
    if (!unzip.unzip(data, logFile, agent))
        return false;

    if (response) {
        agent->setMessage(__DATE__ + std::string(" - ERROR: Failed ") +
                    std::string("to download the agent."));

        *logFile << agent->getMessage() << std::endl;
        (*data)["error"]["failed_to_download_agent"] = agent->getMessage();

        return false;
    }

    return true;
}
