/* Copyright (C) 2017 Centro de Computacao Cientifica e Software Livre
 * Departamento de Informatica - Universidade Federal do Parana - C3SL/UFPR
 *
 * This file is part of simmc-agent
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301,
 * USA.
 */
#include <agent/linux/unzip_agent.h>
#include <string>

/*---------------------------------------------------------------------
 * set Functions
 */

void Unzip::setError(int i) {
    this->error = i;
}

void Unzip::setUF() {
    this->uf = unzOpen(this->zipFileName);
}

void Unzip::setPathToUnzipAgent(std::string p) {
    this->path = p;
}

void Unzip::setPathToUnzipAgent(std::string p, int tot, std::string plus) {
     int i = p.size(), count = 0;
    while (i-->-1)
        if (p[i] == '/')
            if (++count == tot)
                break;
    this->path = p.substr(0, i) + std::string(plus);
}

void Unzip::setZipFileName(const char* name) {
    this->zipFileName = name;
}

/*---------------------------------------------------------------------
 * get Functions
 */

int Unzip::getError() const {
    return this->error;
}

unzFile Unzip::getUF() const {
    return this->uf;
}

std::string Unzip::getPathToUnzipAgent() const {
    return this->path;
}

const char* Unzip::getZipFileName() {
    return this->zipFileName;
}

/*---------------------------------------------------------------------
 * isSet Functions
 */

bool Unzip::UFIsSet() {
    if (this->uf != NULL)
        return true;
    return false;
}

bool Unzip::zipFileNameIsSet() {
    if (this->zipFileName != NULL)
        return true;
    return false;
}

/*---------------------------------------------------------------------
 * changeFileDate method
 * Sets a file's date and time in dos format 
 */

void Unzip::changeFileDate(const char *path, uint32_t dos_date) {
    struct utimbuf ut;
    struct tm newdate;

    this->dosDateToTm(dos_date, &newdate);

    ut.actime = ut.modtime = mktime(&newdate);
    utime(path, &ut);
}

/*---------------------------------------------------------------------
 * dosDateToTm method
 * Convert dos date/time format to struct tm
 */

int Unzip::dosDateToTm(uint64_t dos_date, struct tm *ptm) {
    uint64_t date = (uint64_t)(dos_date >> 16);

    ptm->tm_mday = (uint16_t)(date & 0x1f);
    ptm->tm_mon = (uint16_t)(((date & 0x1E0) / 0x20) - 1);
    ptm->tm_year = (uint16_t)(((date & 0x0FE00) / 0x0200) + 1980);
    ptm->tm_hour = (uint16_t)((dos_date & 0xF800) / 0x800);
    ptm->tm_min = (uint16_t)((dos_date & 0x7E0) / 0x20);
    ptm->tm_sec = (uint16_t)(2 * (dos_date & 0x1f));
    ptm->tm_isdst = -1;

#define datevalue_in_range(min, max, value) ((min) <= (value) && (value) <= (max)) // NOLINT [whitespace/line_length]
    if (!datevalue_in_range(0, 11, ptm->tm_mon) ||
        !datevalue_in_range(1, 31, ptm->tm_mday) ||
        !datevalue_in_range(0, 23, ptm->tm_hour) ||
        !datevalue_in_range(0, 59, ptm->tm_min) ||
        !datevalue_in_range(0, 59, ptm->tm_sec)) {
        /* Invalid date stored, so don't return it. */
        memset(ptm, 0, sizeof(struct tm));
        return -1;
    }

#undef datevalue_in_range
    return 0;
}

int Unzip::makeDir(const char* newdir, Json::Value* data,
                    std::fstream* logFile, Agent* agent) {
    char *buffer = NULL;
    char *p = NULL;
    int len = (int)strlen(newdir);  // NOLINT [readability/casting]

    if (len <= 0)
        return 0;

    buffer = (char*)malloc(len + 1);  // NOLINT [readability/casting]
    if (buffer == NULL) {
        agent->setMessage(__DATE__ +
                std::string(" - ERROR: Cannot Allocating memory."));
        *logFile << agent->getMessage() << std::endl;
        (*data)["error"]["cannot_allocating_memory"] = agent->getMessage();
        return -1;
    }

    snprintf(buffer, sizeof(buffer), newdir);

    if (buffer[len - 1] == '/')
        buffer[len - 1] = 0;

    if (mkdir(buffer, 0700) == 0) {
        free(buffer);
        return 1;
    }

    p = buffer + 1;
    while (1) {
        char hold;
        while (*p && *p != '\\' && *p != '/')
            p++;
        hold = *p;
        *p = 0;

        if ((mkdir(buffer, 0700) == -1) && (errno == ENOENT)) {
            agent->setMessage(__DATE__ +
                std::string(" - ERROR: Couldn't create directory."));
            *logFile << agent->getMessage() << std::endl;
            (*data)["error"]["could_not_create_directory"] =
                                    agent->getMessage();
            free(buffer);
            return 0;
        }

        if (hold == 0)
            break;

        *p++ = hold;
    }

    free(buffer);
    return 1;
}

/*---------------------------------------------------------------------
 * extracts Functions
 */

bool Unzip::unzip(Json::Value* data, std::fstream* logFile, Agent* agent) {
    if (!this->zipFileNameIsSet()) {
        agent->setMessage(__DATE__ +
                std::string(" - ERROR: Zip File name not setted."));
        *logFile << agent->getMessage() << std::endl;
        (*data)["error"]["zip_file_name_not_settted"] = agent->getMessage();
        return false;
    }
    setUF();

    if (!UFIsSet()) {
        agent->setMessage(__DATE__ + std::string(" - ERROR: Cannot open") +
                        std::string(" or doesn't exist zip file."));
        *logFile << agent->getMessage()  << std::endl;
        (*data)["error"]["cannot_open_zip_file"] = agent->getMessage();
    }

    if (this->unzipAllFiles(data, logFile, agent)) {
        this->setZipFileName((agent->getDirInstall() + std::string("/")
                 + agent->getZipName()).c_str());
        boost::filesystem::remove_all(this->getZipFileName());
        agent->setMessage(__DATE__ +
                std::string(" - SUCCESS: Agent extracted successfully."));
        *logFile << agent->getMessage()  << std::endl;
        (*data)["success"]["agent_extracted"] = agent->getMessage();
        return true;
    } else {
        agent->setMessage(__DATE__ +
                std::string(" - ERROR: Failed to extract agent."));
        *logFile << agent->getMessage()  << std::endl;
        (*data)["error"]["failed_extract_agent"] = agent->getMessage();
        return false;
    }
}

bool Unzip::unzipAllFiles(Json::Value* data, std::fstream* logFile,
                            Agent* agent) {
    this->setError(unzGoToFirstFile(this->uf));
    int cont = 1;

    if (this->getError() != UNZ_OK) {
        agent->setMessage(__DATE__ +
                std::string(" - ERROR: Cannot open first file."));
        *logFile << agent->getMessage()  << std::endl;
        (*data)["error"]["cannot_open_first_file"] = agent->getMessage();
        return false;
    }

    do {
        this->setError(unzipCurrentFile(data, logFile, agent, cont));
        if (this->getError()  != UNZ_OK)
            return false;
        this->setError(unzGoToNextFile(this->uf));
        cont++;
    } while (this->getError()  == UNZ_OK);

    if (this->getError()  != UNZ_END_OF_LIST_OF_FILE) {
        agent->setMessage(__DATE__ +
                std::string(" - ERROR: Cannot open the next file."));
        *logFile << agent->getMessage()  << std::endl;
        (*data)["error"]["cannot_open_next_file"] = agent->getMessage();
        return false;
    }



    return true;
}

int Unzip::unzipCurrentFile(Json::Value* data, std::fstream* logFile,
                            Agent* agent, int count) {
    this->setError(UNZ_OK);
    unz_file_info file_info = {0};
    char filename_inzip[256] = {0};
    char *filename_withoutpath = NULL;
    const char *write_filename = NULL;
    FILE* fout = NULL;
    char *p = NULL;
    void* buf = NULL;
    int errclose = UNZ_OK;

    this->setError(unzGetCurrentFileInfo(this->uf, &file_info, filename_inzip,
                    sizeof(filename_inzip), NULL, 0, NULL, 0));

    if (this->getError() != UNZ_OK) {
        agent->setMessage(__DATE__ +
                std::string(" - ERROR: Cannot not read file info."));
        *logFile << agent->getMessage()  << std::endl;
        (*data)["error"]["cannot_read_file_info"] = agent->getMessage();
        return this->getError();
    }

    this->setPathToUnzipAgent(agent->getDirInstall() + std::string("/") +
                                    std::string(filename_inzip));

    p = filename_withoutpath = filename_inzip;
    while (*p != 0) {
        if ((*p == '/') || (*p == '\\'))
            filename_withoutpath = p+1;
        p++;
    }

    /* If zip entry is a directory then create it on disk */
    if (*filename_withoutpath == 0) {
        agent->setMessage(__DATE__ +
                std::string(" - WARNING: Create directory ..."));
        *logFile << agent->getMessage()  << std::endl;
        (*data)["warning"]["create_directory"] = agent->getMessage();

        boost::filesystem::path dir(this->getPathToUnzipAgent().c_str());
        boost::filesystem::create_directory(dir);
        return this->getError();
    }

    buf = (void*)malloc(file_info.uncompressed_size); // NOLINT [readability/casting]
    if (buf == NULL) {
        agent->setMessage(__DATE__ +
                std::string(" - ERROR: Cannot allocatin memory"));
        *logFile << agent->getMessage()  << std::endl;
        (*data)["error"]["cannot_allocating_memory"] = agent->getMessage();
        return UNZ_INTERNALERROR;
    }

    // write_filename = filename_inzip;
    write_filename = this->getPathToUnzipAgent().c_str();

    fout = fopen(write_filename, "wb");
    if (this->getError() == UNZ_OK) {
        this->setError(unzOpenCurrentFile(this->uf));
        if (this->getError() == UNZ_OK) {

            do {
                this->setError(unzReadCurrentFile(this->uf, buf,
                                file_info.uncompressed_size));

                if (this->getError() < 0) {
                    agent->setMessage(__DATE__ +
                    std::string(" - ERROR: ") +
                    std::string(" With zipfile in unzReadCurrentFile -"));
                    *logFile << agent->getMessage()  << std::endl;
                    (*data)["error"]["with_zipfile_1"] = agent->getMessage();
                    break;
                }

                if (this->getError() == 0)
                    break;

                if (fwrite(buf, this->getError(), 1, fout) != 1) {
                    agent->setMessage(__DATE__ +
                    std::string(" - ERROR: Cannot writing extracted file."));
                    *logFile << agent->getMessage()  << std::endl;
                    (*data)["error"]["cannot_allocating_memory"] =
                                        agent->getMessage();
                    this->setError(UNZ_ERRNO);
                    break;
                }
            } while (this->getError() > 0);

            if (fout)
                fclose(fout);

            if (this->getError() == 0)
                this->changeFileDate(write_filename, file_info.dos_date);
        } else {
            agent->setMessage(__DATE__ +
                std::string(" - ERROR: Cannot open file."));
            *logFile << agent->getMessage()  << std::endl;
            (*data)["error"]["cannot_open_file"] = agent->getMessage();
        }
    }

    errclose = unzCloseCurrentFile(uf);
    if (errclose != UNZ_OK) {
        agent->setMessage(__DATE__ +
                std::string(" - ERROR: ") +
                std::string(" With zipfile in unzReadCurrentFile *"));

                *logFile << agent->getMessage()  << std::endl;
                (*data)["error"]["with_zipfile_2"] = agent->getMessage();
    }

    free(buf);
    return this->getError();
}
